#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/power.h>
#include <avr/sleep.h>
#include <util/atomic.h>
#include <util/delay.h>
#include <stdint.h>
#include <stdlib.h>
#include "adc.h"
#include "inout.h"
#include "led.h"
#include "main.h"
#include "uart.h"
#include "time.h"

/*

SSLAC - Simple solar lead acid charge controller
2012-03-28, Frank Bättermann

TODO: Internal temperature always reads at highest possible value !?

Fuses: select internal 128 kHz oscillator and divide clock by 8 (others to default)
Keep in mind that you will have to set ISP speed lower than 4 kHz if you use 16 kHz system clock.
*/

#ifdef DEBUG_UART
char uartBuffer[16];
uint32_t uartLastTime = UINT32_MAX;
#endif

uint16_t tempIntADC = 0;
uint16_t tempInt = 0; // [°C * 100]
uint16_t panelVoltage = 0; // [mV]
uint16_t panelADC = 0;
uint16_t batteryVoltage = 0; // [mV]
uint16_t batteryADC = 0;


typedef struct
{
    uint16_t adc;
    uint16_t value;
} linearizationPointU16_t;

typedef struct
{
    uint8_t size;
    linearizationPointU16_t point[];
} const linearizationTableU16_t;

// values from 2012-04-14 (Peaktech 3360)
linearizationTableU16_t linListBattVoltage =
{
    6,
    {
        {  0,      0},
        {929,   3240},
        {1781,  6166},
        {2616,  9038},
        {3522, 12154},
        {4064, 14021},
    }
};

// values from 2012-04-14 (Peaktech 3360)
linearizationTableU16_t linListPanelVoltage =
{
    8,
    {
        {   0,     0},
        { 551,  3239},
        {1046,  6163},
        {1539,  9036},
        {2076, 12151},
        {2636, 15416},
        {3131, 18292},
        {3670, 21430},
    }
};

// from atmega48 datasheet, in °Kelvin (0 °C = 273.15 °K)
linearizationTableU16_t linListTempInt =
{
    3,
    {
        {225, 22815}, // -45 °C
        {292, 29815}, // 25 °C
        {354, 35815}  // 85 °C
    }
};



// Convert ADC measurement to a useful value (mapped by a linearization table)
// Attention: This function does not use absolute 0 and uint16_t maximum. It
// interpolates the value according to the gradient betweeb the two closest points known.
uint16_t linearizeU16(const linearizationTableU16_t* list, uint16_t adc)
{
    uint32_t gradient;
    uint16_t point_adc;
    uint16_t point_value;
    

    // shortcuts for more speed
    point_adc = list->point[0].adc;
    point_value = list->point[0].value;
    if (adc == point_adc)
        return point_value;
    else if (adc < point_adc)
    {
        // multiply by 1024 for accuracy reasons (compiler converts this to fast shift operations)
        gradient = (uint32_t)(list->point[1].value - point_value) * 1024 / (list->point[1].adc - point_adc);
        return (uint16_t)(((uint32_t)point_value * 1024 - gradient * (point_adc - adc)) / 1024);
    }
    
    // if we got here the adc value is greater than the first one in the table
    for (uint8_t p = 1; p < list->size; p++)
    {
        point_adc = list->point[p].adc;
        point_value = list->point[p].value;
        
        if (adc == point_adc)
            return point_value;
        else if (adc < point_adc)
        {
            uint8_t lastp = p - 1;
            gradient = (uint32_t)(point_value - list->point[lastp].value) * 1024 / (point_adc - list->point[lastp].adc);
            return (uint16_t)(((uint32_t)point_value * 1024 - gradient * (point_adc - adc)) / 1024);
        }
    }
    
    // if got here the adc value is greater than any given in the table
    uint8_t i_last = list->size - 1;
    uint8_t i_prev = i_last - 1;
    gradient = (uint32_t)(list->point[i_last].value - list->point[i_prev].value) * 1024 / (list->point[i_last].adc - list->point[i_prev].adc);
    return (uint16_t)(((uint32_t)list->point[i_last].value * 1024 + gradient * (adc - list->point[i_last].adc)) / 1024);
}


// Read battery voltage and solar panel voltage
void readVoltages(void)
{
    uint16_t adc;
    
    adc_enable();
    
    // ADC0 = panel voltage (voltage divider: 10k - 475 --> factor 22.05)
    adc_setChannel(0);
    adc = 0;
    for (uint8_t i = 0; i < ADC_OVERSAMPLING; i++)
        adc += adc_singleConversion();
    panelADC = adc / ADC_OVERSAMPLING_DIV;
    panelVoltage = linearizeU16(&linListPanelVoltage, panelADC);
   
    // ADC1 = battery voltage (voltage divider: 10k - 825 --> factor 13.12)
    adc_setChannel(1);
    adc = 0;
    for (uint8_t i = 0; i < ADC_OVERSAMPLING; i++)
        adc += adc_singleConversion();
    batteryADC = adc / ADC_OVERSAMPLING_DIV;
    batteryVoltage = linearizeU16(&linListBattVoltage, batteryADC);
    
    // ADC8 = internal temperature
    adc_setChannel(8);
    // datasheet says: ... can be measured by single conversions...
    tempIntADC = adc_singleConversion();
    tempInt = linearizeU16(&linListTempInt, tempIntADC);
    
    //adc_disable();
}


#ifdef DEBUG_UART
void debug_write_uart(void)
{
    // create a copy of the system uptime
    uint16_t uptimeMs;
    uint32_t uptime;
    ATOMIC_BLOCK(ATOMIC_FORCEON)
    {
        uptimeMs = timeMs / TIME_MULTIPLIER;
        uptime = time;
    }
            
    if (uptime < uartLastTime)
        // there was an overflow of uptime in between --> also overflow lastTime
        uartLastTime += DEBUG_UARD_DATA_INTERVAL;
    if (uptime - uartLastTime >= DEBUG_UARD_DATA_INTERVAL)
    {
        uartLastTime = uptime;

        // write data we know to uart
        uart_puts(itoa(uptime, uartBuffer, 10));
        uart_putc('.');
        if (uptimeMs < 10)
            uart_putc('0');
        else if (uptimeMs < 100)
            uart_putc('0');
        uart_puts(itoa(uptimeMs, uartBuffer, 10));
        
        uart_putc(';');
        uart_puts(itoa(tempIntADC, uartBuffer, 10));
        uart_putc(';');
        uart_puts(itoa((int16_t)((int32_t)tempInt - 27315), uartBuffer, 10));
        
        uart_putc(';');
        uart_puts(itoa(batteryADC, uartBuffer, 10));
        uart_putc(';');
        uart_puts(itoa(batteryVoltage, uartBuffer, 10));
        uart_putc(';');
        uart_puts(itoa(panelADC, uartBuffer, 10));
        uart_putc(';');
        uart_puts(itoa(panelVoltage, uartBuffer, 10));
        uart_puts("\n");
    }
}
#endif

int main(void)
{
    // defaults
    uint8_t loadConnected = 0;
    uint8_t panelConnected = 1;
    uint8_t ledCharge = 0;
    led_voltageStatus_t ledVoltageStatus = led_voltageStatusOff;

    // Save power
    power_spi_disable();
    power_timer0_disable();
    power_timer1_disable();
    power_twi_disable();
#ifndef DEBUG_UART
    power_usart0_disable();
#endif
    set_sleep_mode(SLEEP_MODE_IDLE);//EXT_STANDBY);
    
    time_init(); // timer2 overflow is used to wake from standby.
    led_init();
    inout_init();
    adc_init(adc_voltageReferenceInternal, adc_adjustResultRight, adc_interruptDisabled, adc_autoTriggerDisabled, adc_autoTriggerSourceFreeRunning);  
#ifdef DEBUG_UART
    uart_init();
#ifdef CALIBRATION_MODE
    uart_puts("WARNING: Cannot short panel if U_Batt too high!\n");
#endif
    // data header (CSV-style)
    uart_puts("\n\nTime[s];ADC_TempInt;Temp_Int[°C];ADC_Batt;U_Batt[mV];ADC_Panel;U_Panel[mv]\n");
#endif

    // enable interrupts (needed to wake up from sleep)
    sei();

#ifdef DEBUG_UART

#endif

    // main loop
	for (;;)
	{
        // perform measurements
        readVoltages();
    	
#ifdef DEBUG_UART
        debug_write_uart();
#endif
    	
    	// connect/disconnet load (with hysteresis)
	    if (loadConnected && batteryVoltage < BATT_MIN_VOLTAGE)
	    {
            loadConnected = 0;
            inout_disconnectLoad();
#ifdef DEBUG_UART
            uart_puts("Dropped load.\n");
#endif
	    }
	    else if (!loadConnected && batteryVoltage >= BATT_RECONNECT_LOAD)
	    {
            loadConnected = 1;
            inout_connectLoad();
#ifdef DEBUG_UART
            uart_puts("Connected load.\n");
#endif
	    }
    	
    	// connect/disconnet solar panel (with hysteresis)
	    if (panelConnected && batteryVoltage >= BATT_MAX_VOLTAGE)
	    {


#ifdef CALIBRATION_MODE
            uart_puts("WARNING: Cannot disconnect panel (calibration mode).\n");
#else
            panelConnected = 0;
            inout_disconnectPanel();
#ifdef DEBUG_UART
            uart_puts("Panel disconnected.\n");
#endif
#endif
        }
	    else if (!panelConnected && batteryVoltage <= BATT_RECHARGE)
	    {
            panelConnected = 1;
            inout_connectPanel();
#ifdef DEBUG_UART
            uart_puts("Panel connected.\n");
#endif	    
        }
	    
        // charge led control (with hysteresis)
        if (ledCharge && panelVoltage < batteryVoltage + LED_CHARGE_THRESHOLD - LED_CHARGE_HYSTERESIS / 2)
        {
            ledCharge = 0;
            led_chargingOff();
        }
        if (!ledCharge && panelVoltage >= batteryVoltage + LED_CHARGE_THRESHOLD + LED_CHARGE_HYSTERESIS / 2)
        {
            ledCharge = 1;
            led_chargingOn();
        }

        // battery voltage LED
        if (ledVoltageStatus == led_voltageStatusOff)
        {
            // initial status or or load was dropped
            if (batteryVoltage >= LED_VOLTAGE_HIGH)
                ledVoltageStatus = led_voltageStatusHigh;
            else if (batteryVoltage >= LED_VOLTAGE_MEDIUM)
                ledVoltageStatus = led_voltageStatusMedium;
            else
                ledVoltageStatus = led_voltageStatusLow;
        }
        else if (ledVoltageStatus == led_voltageStatusHigh && batteryVoltage < LED_VOLTAGE_HIGH - LED_VOLTAGE_HYSTERESIS / 2)
            // high --> medium
            ledVoltageStatus = led_voltageStatusMedium;
        else if (ledVoltageStatus == led_voltageStatusMedium && batteryVoltage > LED_VOLTAGE_HIGH + LED_VOLTAGE_HYSTERESIS / 2)
            // medium --> high
            ledVoltageStatus = led_voltageStatusHigh;
        else if (ledVoltageStatus == led_voltageStatusMedium && batteryVoltage < LED_VOLTAGE_MEDIUM - LED_VOLTAGE_HYSTERESIS / 2)
            // medium --> low
            ledVoltageStatus = led_voltageStatusLow;
        else if (ledVoltageStatus == led_voltageStatusLow && batteryVoltage > LED_VOLTAGE_MEDIUM + LED_VOLTAGE_HYSTERESIS / 2)
            // low --> medium
            ledVoltageStatus = led_voltageStatusHigh;
        else if (ledVoltageStatus == led_voltageStatusLow && !loadConnected)
            // low --> off
            ledVoltageStatus = led_voltageStatusOff;
        switch (ledVoltageStatus)
        {
            case led_voltageStatusLow:      led_voltageRed(); break;
            case led_voltageStatusMedium:   led_voltageYellow(); break;
            case led_voltageStatusHigh:     led_voltageGreen(); break;
            default: led_voltageOff();
        }
        
        // sleept until timer2 wakes us
        sleep_mode();
	}

	return 0;
}

